
from dataclasses import dataclass

__all__ = ["Node"]


@dataclass
class Node:
    """This is a node in the network"""
    label: int = None           # The node's label
    begin_to: int = None        # where to links begin in link list
    end_to: int = None			# how many to links in link list
    self_w: int = None

    begin_p: int = None         # play matrix begin and end in link vector
    end_p: int = None
    self_p: int = None

    begin_we: int = None        # weekend begin and end
    end_we: int = None
    self_we: int = None

    day_foi: float = 0.0        # numerator only
    night_foi: float = 0.0      # numerator only
    weekend_foi: float = 0.0    # numerator only

    play_suscept: float = 0.0
    save_play_suscept: float = 0.0

    denominator_n: float = 0.0  #Denominator only
    denominator_d: float = 0.0  # Maybe won't need

    denominator_p: float = 0.0
    denominator_pd: float = 0.0
    x: float = 0.0
    y: float = 0.0
    b: float = 0.0

    id: str = None
    vacid: int = None



"""
Utility functions that are useful for the program but that
should not be immediately visible or used by the user.

Generally, interaction with the program should be via
top-level functions and classes
"""

from ._initialise_infections import *
from ._read_done_file import *

# pyx imports
from ._build_wards_network import *
from ._add_wards_network_distance import *
from ._get_min_max_distances import *
from ._reset_everything import *
from ._rescale_matrix import *
from ._recalculate_denominators import *
from ._move_population import *
from ._fill_in_gaps import *
from ._build_play_matrix import *
from ._workspace import *
from ._run_model import *
from ._array import *
from ._iterate import *
from ._iterate_weekend import *
from ._import_infection import *
from ._ran_binomial import *
from ._parallel import *
from ._assert_sane_network import *




from dataclasses import dataclass

__all__ = ["Population"]


@dataclass
class Population:
    """This class holds information about the progress of the
       disease through the population
    """
    initial: int = 0        # the initial population loaded into the model
    susceptibles: int = 0   # the number of members who could be infected
    latent: int = 0         # the number of latent infections
    total: int = 0          # the total number of infections
    recovereds: int = 0     # the total number of recovered members
    n_inf_wards: int = 0    # the number of infected in wards

    @property
    def population(self) -> int:
        return self.susceptibles + self.total + self.recovereds

    def __str__(self):
        return f"S: {self.susceptibles}    " \
               f"E: {self.latent}    " \
               f"I: {self.total}    " \
               f"R: {self.recovereds}    " \
               f"IW: {self.n_inf_wards}   " \
               f"TOTAL POPULATION {self.population}"



#!/bin/env/python3
#cython: boundscheck=False
#cython: cdivision=True
#cython: initializedcheck=False
#cython: cdivision_warnings=False
#cython: wraparound=False
#cython: binding=False
#cython: initializedcheck=False
#cython: nonecheck=False
#cython: overflowcheck=False

from libc.math cimport exp

cdef inline double rate_to_prob(double rate) nogil:
    """Convert the return the probability associated with the passed
       infection rate
    """
    return rate - (0.5*rate*rate) if rate < 1e-6 else 1.0 - exp(-rate)




from dataclasses import dataclass
import os
import pathlib

__all__ = ["InputFiles"]

_inputfiles = {}

_default_model_path = os.path.join(pathlib.Path.home(),
                                   "GitHub", "MetaWardsData")

_default_folder_name = "model_data"


@dataclass
class InputFiles:
    work: str = None                # WORK MATRIX
    play: str = None                # PLAY MATRIX
    identifier: str = None          # WARD NAMES
    identifier2: str = None         # WARD IDs (Communities, Counties,
                                    #           Districts, UA's etc)
    weekend: str = None             # WEEKEND MATRIX
    play_size: str = None           # SIZE OF POPULATION IN THE PLAY PILE
    position: str = None            # CENTRE of BOUNDING BOXES
    seed: str = None                # LIST OF SEED NODES
    nodes_to_track: str = None      # LIST OF NODES TO TRACK
    additional_seeding: str = None  # LIST OF EXTRA SEED WARDS...
    uv: str = None                  # UV file

    _filename: str = None            # Full path of the description.json file
    _model_name: str = None          # Name of the model
    _model_path: str = None          # Directory containing the input files
    _model_version: str = None       # Version loaded from the data
    _authors: str = None             # Author(s) of this data
    _contacts: str = None            # Contact(s) for this data
    _references: str = None          # References for this data
    _repository: str = None          # GitHub repository for this data
    _repository_version: str = None  # Git version of the data

    def model_name(self):
        """Return the name of this model"""
        return self._model_name

    def model_path(self):
        """Return the path to the directory containing this model"""
        return self._model_path

    def model_version(self):
        """Return the version of the data in this model"""
        return self._model_version

    def __str__(self):
        return f"Model {self._model_name} version {self._model_version}\n" \
               f"loaded from {self._filename}\n" \
               f"root directory {self._model_path}\n" \
               f"author(s): {self._authors}\n" \
               f"contact(s): {self._contacts}\n" \
               f"references(s): {self._references}\n" \
               f"repository: {self._repository}\n" \
               f"repository_version: {self._repository_version}\n\n" \
               f"work = {self.work}\n" \
               f"play = {self.play}\n" \
               f"identifier = {self.identifier}\n" \
               f"identifier2 = {self.identifier2}\n" \
               f"weekend = {self.weekend}\n" \
               f"play_size = {self.play_size}\n" \
               f"position = {self.position}\n" \
               f"seed = {self.seed}\n" \
               f"nodes_to_track = {self.nodes_to_track}\n" \
               f"additional_seeding = {self.additional_seeding}\n\n"

    def _localise(self):
        """Localise the filenames in this input files set. This will
           prepend model_path/model to every filename and will also
           double-check that all files exist and are readable
        """
        members = [attr for attr in dir(self)
                    if not callable(getattr(self, attr))
                        and not attr.startswith("_")]

        for member in members:
            filename = getattr(self, member)
            if filename:
                filename = os.path.join(self._model_path, filename)

                if not (os.path.exists(filename) and os.path.isfile(filename)):
                    raise FileNotFoundError(
                            f"Cannot find input file {member} = {filename}")

                setattr(self, member, filename)

    @staticmethod
    def load(model: str = "2011Data",
             repository: str=_default_model_path,
             folder: str=_default_folder_name,
             description: str="description.json",
             filename: str = None):
        """Load the parameters associated with the passed model.
           This will look for the parameters specified in
           the json file called f"{repository}/{folder}/{model}/{description}"

           By default this will load the 2011Data parameters
           from $HOME/GitHub/model_data/2011Data/description.json

           Alternatively you can provide the full path to the
           description json file usng the 'filename' argument.
           All files within this description will be searched
           for using the directory that contains that file
           as a base
        """
        repository_version = None

        if filename is None:
            if repository is None:
                repository = os.getenv("METAWARDSDATA")
                if repository is None:
                    repository = _default_model_path

            from ._parameters import get_repository_version
            repository_version = get_repository_version(repository)
            filename = os.path.join(repository, folder,
                                    model, description)

        json_file = filename
        model_path = os.path.dirname(filename)

        try:
            with open(json_file, "r") as FILE:
                import json
                files = json.load(FILE)

        except Exception as e:
            print(f"Could not find the model file {json_file}")
            print(f"Either it does not exist of was corrupted.")
            print(f"Error was {e.__class__} {e}")
            print(f"To download the model data type the command:")
            print(f"  git clone https://github.com/metawards/MetaWardsData")
            print(f"and then re-run this function passing in the full")
            print(f"path to where you downloaded this directory")
            raise FileNotFoundError(f"Could not find or read {json_file}: "
                                    f"{e.__class__} {e}")

        model = InputFiles(work=files["work"],
                           play=files["play"],
                           identifier=files["identifier"],
                           identifier2=files["identifier2"],
                           weekend=files["weekend"],
                           play_size=files["play_size"],
                           position=files["position"],
                           seed=files["seed"],
                           nodes_to_track=files["nodes_to_track"],
                           additional_seeding=files["additional_seeding"],
                           uv=files["uv"],
                           _filename=json_file,
                           _model_path=model_path,
                           _model_name=files["name"],
                           _model_version=files["version"],
                           _references=files["reference(s)"],
                           _authors=files["author(s)"],
                           _contacts=files["contact(s)"],
                           _repository=repository,
                           _repository_version=repository_version)

        model._localise()

        return model



#!/bin/env python3

# This is the command-line version of metawards

def cli():
    import sys
    import argparse

    parser = argparse.ArgumentParser(
                    description="MetaWards epidemic modelling - see "
                                "https://github.com/metawards/metawards "
                                "for more information",
                    prog="metawards")

    parser.add_argument('-i', '--input', type=str,
                        help="Input file for the simulation")

    parser.add_argument('-l', '--line', type=int, default=0,
                        help="Line number from the inputfile to run "
                             "(default 0 as 0-indexed line number)")

    parser.add_argument('-s', '--seed', type=int, default=None,
                        help="Random number seed for this run "
                             "(default is to use a random seed)")

    parser.add_argument('-u', '--UV', type=float, default=1.0,
                        help="Value for the UV parameter for the model "
                             "(default is 1.0)")

    parser.add_argument('-d', '--disease', type=str, default="ncov",
                        help="Name of the disease to model "
                             "(default is 'ncov')")

    parser.add_argument('-I', '--input-data', type=str, default="2011Data",
                        help="Name of the input data set for the network "
                             "(default is '2011Data')")

    parser.add_argument('-p', '--parameters', type=str, default="march29",
                        help="Name of the input parameter set used to "
                             "control the simulation (default 'march29')")

    parser.add_argument('-r', '--repository', type=str, default=None,
                        help="Path to the MetaWardsData repository. If "
                             "unspecified this defaults to the value "
                             "in the environment variable METAWARDSDATA "
                             "or, if that isn't specified, to "
                             "$HOME/GitHub/MetaWardsData")

    parser.add_argument('-P', '--population', type=int, default=57104043,
                        help="Initial population (default 57104043)")

    parser.add_argument('-n', '--nsteps', type=int, default=None,
                        help="Maximum number of steps to run for the "
                             "simulation (default is to run until the "
                             "epidemic has finished)")

    parser.add_argument('--nthreads', type=int, default=None,
                        help="Number of threads over which to run the "
                             "model.")

    parser.add_argument('-o', '--output', type=str, default="output",
                        help="Path to the directory in which to place all "
                             "output files (default 'output')")

    parser.add_argument('--profile', action="store_true",
                        default=True, help="Enable profiling of the code")

    parser.add_argument('--no-profile', action="store_true",
                        default=False, help="Disable profiling of the code")

    args = parser.parse_args()

    if args.input is None:
        parser.print_help(sys.stdout)
        sys.exit(0)

    # import the parameters here to speed up the display of help
    from metawards import Parameters, Network, Population

    # load all of the parameters
    try:
        params = Parameters.load(parameters=args.parameters)
    except Exception as e:
        print(f"Unable to load parameter files. Make sure that you have "
              f"cloned the MetaWardsData repository and have set the "
              f"environment variable METAWARDSDATA to point to the "
              f"local directory containing the repository, e.g. the "
              f"default is $HOME/GitHub/MetaWardsData")
        raise e

    # should we profile the code? (default yes, as it doesn't cost anything)
    profile = args.profile

    if args.no_profile:
        profile = False

    # load the disease and starting-point input files
    params.set_disease(args.disease)
    params.set_input_files(args.input_data)

    # start from the parameters in the specified line number of the
    # provided input file
    params.read_file(args.input, args.line)

    # extra parameters that are set
    params.UV = args.UV

    # set these extra parameters to 0
    params.static_play_at_home = 0
    params.play_to_work = 0
    params.work_to_play = 0
    params.daily_imports = 0.0

    # the size of the starting population
    population = Population(initial=args.population)

    print("Building the network...")
    network = Network.build(params=params, calculate_distances=True,
                            profile=profile)

    print("Run the model...")
    population = network.run(population=population, seed=args.seed,
                             s=-1, nsteps=args.nsteps,
                             output_dir=args.output,
                             profile=profile,
                             nthreads=args.nthreads
                             )

    print("End of the run")

    print(f"Model output:  {population}")


if __name__ == "__main__":
    cli()




__all__ = ["clear_all_infections"]


def clear_all_infections(infections, play_infections):
    """Clears all infections (held in the list of array(int) in infections,
       and array(int) in play_infections) associated with the
       passed network
    """
    assert len(infections) == len(play_infections)

    for i in range(0, len(infections)):
        a = infections[i]
        for j in range(0, len(a)):
            a[j] = 0

        a = play_infections[i]
        for j in range(0, len(a)):
            a[j] = 0



# Package setup copied from
# https://github.com/FedericoStra/cython-package-example
# Thanks - this was really helpful :-)

import os
import versioneer
from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize

cflags = '-O3 -march=native -Wall -fopenmp'

random_sources = ["src/metawards/ran_binomial/mt19937.c", 
                  "src/metawards/ran_binomial/distributions.c"]

# https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#distributing-cython-modules
def no_cythonize(extensions, **_ignore):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions

# Currently each file compiles to its own module. Ideally I'd like
# them all to compile into a single module library, but this doesn't
# work (get the error "ImportError: dynamic module does not define
# init function" and the answers on stackoverflow didn't work.
# If anyone can help please do :-)

extensions = [
    Extension("metawards._nodes", ["src/metawards/_nodes.pyx"]),
    Extension("metawards._links", ["src/metawards/_links.pyx"]),
    Extension("metawards._run_model", ["src/metawards/_run_model.pyx"]),
    Extension("metawards._workspace", ["src/metawards/_workspace.pyx"]),
    Extension("metawards._build_wards_network", ["src/metawards/_build_wards_network.pyx"]),
    Extension("metawards._add_wards_network_distance", ["src/metawards/_add_wards_network_distance.pyx"]),
    Extension("metawards._get_min_max_distances", ["src/metawards/_get_min_max_distances.pyx"]),
    Extension("metawards._reset_everything", ["src/metawards/_reset_everything.pyx"]),
    Extension("metawards._rescale_matrix", ["src/metawards/_rescale_matrix.pyx"]),
    Extension("metawards._recalculate_denominators", ["src/metawards/_recalculate_denominators.pyx"]),
    Extension("metawards._move_population", ["src/metawards/_move_population.pyx"]),
    Extension("metawards._fill_in_gaps", ["src/metawards/_fill_in_gaps.pyx"]),
    Extension("metawards._build_play_matrix", ["src/metawards/_build_play_matrix.pyx"]),
    Extension("metawards._array", ["src/metawards/_array.pyx"]),
    Extension("metawards._iterate", ["src/metawards/_iterate.pyx"]+random_sources),
    Extension("metawards._iterate_weekend", ["src/metawards/_iterate_weekend.pyx"]),
    Extension("metawards._extract_data", ["src/metawards/_extract_data.pyx"]),
    Extension("metawards._extract_data_for_graphics", ["src/metawards/_extract_data_for_graphics.pyx"]),
    Extension("metawards._import_infection", ["src/metawards/_import_infection.pyx"]+random_sources),
    Extension("metawards._ran_binomial", ["src/metawards/_ran_binomial.pyx"]+random_sources),
    Extension("metawards._assert_sane_network", ["src/metawards/_assert_sane_network.pyx"]),
    Extension("metawards._parallel", ["src/metawards/_parallel.pyx"]),
]

# disable bounds checking and wraparound globally as the code already
# checks for this and doesn't use negative indexing. It can be 
# turned on as needed in the modules
for e in extensions:
    e.cython_directives = {"boundscheck": False, "wraparound": False}

CYTHONIZE = bool(int(os.getenv("CYTHONIZE", 0)))

os.environ['CFLAGS'] = cflags

if CYTHONIZE:
    compiler_directives = {"language_level": 3, "embedsignature": True}
    extensions = cythonize(extensions, compiler_directives=compiler_directives)
else:
    extensions = no_cythonize(extensions)

with open("requirements.txt") as fp:
    install_requires = fp.read().strip().split("\n")

with open("requirements-dev.txt") as fp:
    dev_requires = fp.read().strip().split("\n")

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    ext_modules=extensions,
    install_requires=install_requires,
    extras_require={
        "dev": dev_requires,
        "docs": ["sphinx", "sphinx-rtd-theme"]
    },
    entry_points={
        "console_scripts": [
           "metawards = metawards.app.run:cli"
        ]
    }
)
